from numpy import array, empty, atleast_1d, log, exp
from scipy.integrate import quad
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures
from ct2vl.rates import make_rates


class Converter:
    """Uses PCR reaction curves to calibrate a model which converts Ct values to viral loads

    Parameters
    ----------
    traces: str, pandas.DataFrame, numpy.ndarray, or dict
        More commonly, a table where each row corresponds to a PCR reaction curve and
        each column is a cycle in the reaction.
        (OR, if this contains columns named max_replication_rate and max_replication_rate_cycle,
        these columns are parallel arrays where for trace[j], the maximum replication rate was
        max_replication_rate[j] and that rate was observed at cycle max_replication_rate_cycle[j].
        In the unlikely case that this is the type of input data you have.)
    LoD: float
        Limit of detection (LoD): copies of SARS-CoV-2 viral genomes/mL (copies/mL; viral load at the LoD)
    Ct_at_LoD
        Ct value at the limit of detection (LoD)
    """

    """
    Member variables
    ----------------
    LoD: float
    Ct_at_LoD: float
    max_replication_rate_cycle: array = field(init=False)
    max_replication_rate: array = field(init=False)
    model: LinearRegression = field(init=False)
    """

    def __init__(self, traces, LoD, Ct_at_LoD):
        self.replication_rate_supplier = make_rates(traces)
        self.calibrate()
        self.LoD = LoD
        self.Ct_at_LoD = Ct_at_LoD

    def calibrate(self):
        """Fits a polynomial linear regression, where the degree of the polynomial is chosen via
        5-fold grid search cross-validation.
        """
        pipeline = make_pipeline(
            PolynomialFeatures(), LinearRegression(fit_intercept=False)
        )
        cv = GridSearchCV(pipeline, {"polynomialfeatures__degree": [1, 2, 3]})
        cv.fit(
            X=self.replication_rate_supplier.max_replication_rate_cycle,
            y=self.replication_rate_supplier.max_replication_rate,
        )
        self.model = cv.best_estimator_

    def log_replication_rate(self, Ct):
        """Predicts the log replication rate for a give Ct value using a model fit"""
        return log(self.model.predict(array([[Ct]])))

    def ct_to_viral_load(self, Ct):
        """Converts Ct values to viral loads
        Parameters
        ----------
        Ct: Iterable[float] or float
        """
        Ct = atleast_1d(Ct)
        viral_loads = empty(Ct.shape)
        for i, ct_i in enumerate(Ct):
            integral_Ct, _ = quad(self.log_replication_rate, 0, ct_i)
            integral_Ct_at_LoD, _ = quad(self.log_replication_rate, 0, self.Ct_at_LoD)
            viral_loads[i] = exp(log(self.LoD) + integral_Ct_at_LoD - integral_Ct)
        return viral_loads



from abc import ABC
from numpy import ndarray
from pandas import DataFrame, read_csv


def make_rates(data):
    """Instantiates a concrete subclass of IReplicationRateSupplier.

    Parameters
    ----------
    data: Something that can be converted into a DataFrame.
       The DataFrame will either contain arrays for
       max_replication_rate and max_replication_rate_cycle,
       or will be the data from a traces file.
       A trace file shall contain a table where each row corresponds
       to a PCR reaction curve and
       each column is a cycle in the reaction.
    """
    options = {
        str: read_csv,
        DataFrame: lambda df: df,
        dict: lambda d: DataFrame.from_dict(d),
        ndarray: DataFrame,
    }
    data = options[type(data)](data)
    if "max_replication_rate" in data:
        return ReplicationRates(data)
    else:
        return ReplicationRateFromTraces(data)


class IReplicationRateSupplier(ABC):
    """Abstract base class for classes that can supply the max_replication_rate
    and max_replication_rate_cycle arrays.
    Any subclass must have member variables named max_replication_rate
    and max_replication_rate_cycle (or corresponding __get__ methods),
    which return ndarrays."""


class ReplicationRates(IReplicationRateSupplier):
    def __init__(self, data):
        """Converts input to two arrays, giving the
        max_replication_rate and max_replication_rate_cycle values

        Parameters
        __________
        data: pandas.DataFrame
           A table whose columns include max_replication_rate
           and max_replication_rate_cycle
           Note that the keys or column names must be exactly these strings.
        """
        self.max_replication_rate = (
            data["max_replication_rate"].to_numpy().reshape(-1, 1)
        )
        self.max_replication_rate_cycle = (
            data["max_replication_rate_cycle"].to_numpy().reshape(-1, 1)
        )


class ReplicationRateFromTraces(IReplicationRateSupplier):
    def __init__(self, traces):
        self.traces = traces
        self.preprocess_traces()
        self.calc_max_replication_rate()

    def preprocess_traces(self):
        """Preprocesses PRC reaction curves via dropping initial values,
        removing negative values, and making the values monotonic.
        """
        self.traces = self.traces.T
        # Remove first 3 rows, since early values tend to be noise
        self.traces = self.traces.iloc[3:]
        # Negative values are noise, so we can set them to zero.
        self.traces[self.traces < 0] = 0
        # Theoretically, product should only increase, so we can make the data monotonic.
        self.traces = self.traces.cummax()
        # Add a positive constant to prevent division by zero
        self.traces = self.traces + 1

    def calc_max_replication_rate(self):
        """Calculates the ratio between the (i+1)th and ith value of the reaction curve
        then takes the max and argmax of the sequence of ratios.
        """
        # Divide (i+1)th value by the ith value
        replication_rate = self.traces.div(self.traces.shift().bfill())
        self.max_replication_rate_cycle = (
            replication_rate.idxmax().astype(int).to_numpy().reshape(-1, 1)
        )
        self.max_replication_rate = replication_rate.max().to_numpy().reshape(-1, 1)



from argparse import ArgumentParser


def configure_arguments():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(
        dest="mode", help="modes include: calibrate or convert"
    )
    parser_calibrate = subparsers.add_parser("calibrate", help="Calibrates ct2vl")
    parser_calibrate.add_argument(
        "traces",
        type=str,
        help=("Filepath to a csv file containing PCR reaction traces"),
    )
    parser_calibrate.add_argument(
        "LoD",
        type=float,
        help=(
            "Limit of detection (LoD): copies of SARS-CoV-2 viral genomes/mL (copies/mL; viral load at the LoD)"
        ),
    )
    parser_calibrate.add_argument(
        "Ct_at_LoD",
        type=float,
        help=("Ct value at the limit of detection (LoD)"),
    )
    parser_convert = subparsers.add_parser(
        "convert", help="Predicts viral load from given Ct values"
    )
    parser_convert.add_argument(
        "Ct", nargs="+", type=float, help="Ct value to convert to viral load"
    )
    parser_convert.add_argument(
        "-o", "--outfile", type=str, help="Filepath for results"
    )
    return parser.parse_args()



from pickle import dump, load
from numpy import log10
from pandas import DataFrame
from ct2vl.cli import configure_arguments
from ct2vl.conversion import Converter
from pathlib import Path
from os.path import abspath, dirname


def main():
    module_path = Path(abspath(dirname(__file__)))
    filename = Path("calibration.pkl")
    calibration_filepath = module_path / filename
    args = configure_arguments()

    if args.mode == "calibrate":
        converter = Converter(args.traces, args.LoD, args.Ct_at_LoD)
        with open(calibration_filepath, "wb") as f:
            dump(converter, f)
        print("Calibration complete.")
    elif args.mode == "convert":
        if not calibration_filepath.is_file():
            raise ValueError(
                "You must calibrate ct2vl before you can use the 'convert' argument."
            )
        with open(calibration_filepath, "rb") as f:
            calibrated_converter = load(f)
        viral_load = calibrated_converter.ct_to_viral_load(args.Ct)
        log10_viral_load = log10(viral_load)
        formatted_results = DataFrame(
            {
                "LoD": calibrated_converter.LoD,
                "Ct_at_LoD": calibrated_converter.Ct_at_LoD,
                "Ct": args.Ct,
                "viral_load": viral_load,
                "log10_viral_load": log10_viral_load,
            }
        )
        formatted_results.index += 1
        print(formatted_results)
        if args.outfile:
            formatted_results.to_csv(
                args.outfile, sep="\t", float_format="%.3f", index=False
            )


if __name__ == "__main__":
    main()  # pragma: no cover
